#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @FileName     :   func.py
# @Function     :   使用转换函数
from random import choice
from typing import List


def xor(a, b) -> list:
    return list(map(lambda x, y: x ^ y, a, b))


def rotl(x, n):
    return ((x << n) & 0xffffffff) | ((x >> (32 - n)) & 0xffffffff)


def get_uint32_be(key_data):
    return ((key_data[0] << 24) | (key_data[1] << 16) | (key_data[2] << 8) | (key_data[3]))


def put_uint32_be(n):
    return [((n >> 24) & 0xff), ((n >> 16) & 0xff), ((n >> 8) & 0xff), (n & 0xff)]


def padding(data, block=16):
    return data + [(16 - len(data) % block) for _ in range(16 - len(data) % block)]


def unpadding(data):
    return data[:-data[-1]]


def list_to_bytes(data: List[int]) -> bytes:
    return b''.join([bytes((i,)) for i in data])


def bytes_to_list(data: bytes) -> List[int]:
    return [i for i in data]


def random_hex(n: int) -> str:
    return ''.join([choice('0123456789abcdef') for _ in range(n)])


def int_to_hex(num: int) -> str:
    return hex(num)[2:] if len(hex(num)) % 2 == 0 else '0' + hex(num)[2:]


def hex_to_int(hex_str: str) -> int:
    return int(hex_str, 16)


def hex_to_bytes(*args: str) -> bytes:
    return bytes.fromhex(''.join(args))


def get_mask(para_len):
    return int('8' + '0' * (para_len - 1), 16)


def h256(*args) -> bytes:
    """拼接所有16进制字符串 并转为 二进制"""
    return bytes.fromhex(''.join(args))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @FileName     :   ec.py
# @Function     :   椭圆曲线算法
# 参考Crypto实现 https://github.com/ashutosh1206/Crypton/blob/master/Elliptic-Curves/ellipticcurve.py
from typing import Optional

from .utils import hex_to_int


def inverse_mod(a, m):
    if a < 0 or m <= a:
        a = a % m
    c, d = a, m
    uc, vc, ud, vd = 1, 0, 0, 1
    while c != 0:
        q, c, d = divmod(d, c) + (c,)
        uc, vc, ud, vd = ud - q * uc, vd - q * vc, uc, vc
    assert d == 1
    if ud > 0:
        return ud
    else:
        return ud + m


class CurveFp:
    """Fp有限域曲线 方程 y^2= x^3+ax+b"""
    name: str  # 曲线名称
    key_size: int  # 曲线位数

    # 必要参数
    a: str  # 曲线方程 y^2= x^3+ax+b 的系数a, 16进制字符串
    b: str  # 曲线方程 y^2= x^3+ax+b 的系数a, 16进制字符串
    p: str  # 大于3的一个大素数

    gx: str  # 基点x坐标, 16进制字符串
    gy: str  # 基点y坐标, 16进制字符串
    n: str  # 曲线的阶, 如果椭圆曲线上一点P，存在最小的正整数n使得数乘nP=O∞ ,则将n称为P的阶若n不存在，则P是无限阶的

    def params(self) -> dict:
        """
        曲线参数
        :return: 字典类型的曲线参数
        """
        return dict(a=self.a, b=self.b, p=self.p, gx=self.gx, gy=self.gy, n=self.n)

    def is_on_curve(self, x: int, y: int) -> bool:
        """
        点(x, y)是否在曲线上 y^2 - (x^3 + ax + b) 应为p的倍数
        :param x: x坐标
        :param y: y坐标
        :return: 在曲线上返回True, 否则返回False
        """
        # _verify if y^2 - (x^3 + ax + b) is a multiple of p,
        a, b, p = hex_to_int(self.a), hex_to_int(self.b), hex_to_int(self.p)

        # (y^2 - x^3 - a - b ) % p == 0
        return (y ** 2 - x ** 3 - a * x - b) % p == 0


class Point:
    """有限领Fp曲线Curve上的点"""

    def __init__(self, curve: Optional[CurveFp],
                 x: Optional[int], y: Optional[int],
                 order=None):
        self.curve = curve
        self.x = x
        self.y = y
        self.order = order

        if self.curve:
            assert self.curve.is_on_curve(x, y), '点(x, y)不在曲线上'

        if order:
            assert self * order == INFINITY

    def __add__(self, other):
        assert isinstance(other, Point), '仅支持点+点'
        if other == INFINITY:
            return self
        if self == INFINITY:
            return other
        assert self.curve == other.curve, '两个点不在同一条曲线上'

        p = hex_to_int(self.curve.p)  # 16进制字符串转int

        if self.x == other.x:
            if (self.y + other.y) % p == 0:
                return INFINITY
            else:
                return self.double()

        l = ((other.y - self.y) * inverse_mod(other.x - self.x, p)) % p

        x3 = (l * l - self.x - other.x) % p
        y3 = (l * (self.x - x3) - self.y) % p
        return Point(self.curve, x3, y3)

    def __mul__(self, other: int):
        assert isinstance(other, int), '仅支持乘以整数'

        def leftmost_bit(x):
            assert x > 0, 'x不大于0'
            result = 1
            while result <= x:
                result = 2 * result
            return result // 2

        e = other
        if self.order:
            e = e % self.order
        if e == 0:
            return INFINITY
        if self == INFINITY:
            return INFINITY
        assert e > 0, 'e不大于0'

        e3 = 3 * e

        # 相对于x轴的反向点
        negative_self = Point(self.curve, self.x, -self.y, self.order)

        i = leftmost_bit(e3) // 2
        result = self
        while i > 1:
            result = result.double()
            if (e3 & i) != 0 and (e & i) == 0:
                result = result + self
            if (e3 & i) == 0 and (e & i) != 0:
                result = result + negative_self
            i = i // 2
        return result

    def __rmul__(self, other: int):
        """反乘法，当左边对象不支持时调用，如 2 * point """
        assert isinstance(other, int), '仅支持乘以整数'
        return self * other

    def __str__(self):
        if self == INFINITY:
            return "infinity"
        return "(%d,%d)" % (self.x, self.y)

    def __sub__(self, other):
        """减法"""
        assert isinstance(other, Point), '仅支持 点-点 '
        return self + (-other)

    def __neg__(self):
        """负数值，x轴对称点"""
        if self == INFINITY:
            return INFINITY
        return Point(self.curve, self.x, -self.y)

    def __eq__(self, other):
        if not isinstance(other, Point):
            return False
        return self.curve == other.curve and self.x == other.x and self.y == other.y

    def double(self):
        if self == INFINITY:
            return INFINITY

        p = hex_to_int(self.curve.p)
        a = hex_to_int(self.curve.a)

        _l = ((3 * self.x * self.x + a) * inverse_mod(2 * self.y, p)) % p
        x3 = (_l * _l - 2 * self.x) % p
        y3 = (_l * (self.x - x3) - self.y) % p
        return Point(self.curve, x3, y3)


INFINITY = Point(None, None, None)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @FileName     :   sm2.py
# @Function     :   SM2算法实现
"""
签名者用户A的密钥对包括其私钥dA和公钥PA=[dA]G= (xA,yA)
签名者用户A具有长度为entlenA比特的可辨别标识IDA，
ENTLA是由整数entlenA转换而成的两个字节
ZA=h256(ENTLA || IDA || a || b || xG || yG|| xA || yA)。
待签名的消息为M，
数字签名(r,s)
"""
import base64
import binascii
import typing
from typing import Union

from pyasn1.codec.der import decoder, encoder
from pyasn1.type.univ import BitString, Integer, OctetString, SequenceOf

from . import ec, sm2_asn1, sm3
from .utils import h256, hex_to_bytes, hex_to_int, int_to_hex, random_hex

# 默认用户A id
DEFAULT_UID = b'1234567812345678'
PARA_LEN = 64


class SM2P256Curve(ec.CurveFp):
    name = 'sm2p256v1'
    key_size = 256  # 素数域256 位椭圆曲线

    gx = '32C4AE2C1F1981195F9904466A39C9948FE30BBFF2660BE1715A4589334C74C7'  # 基点 Gx, 16进制字符串
    gy = 'BC3736A2F4F6779C59BDCEE36B692153D0A9877CC62A474002DF32E52139F0A0'  # 基点 Gy, 16进制字符串

    # 如果椭圆曲线上一点P，存在最小的正整数n使得数乘nP=O∞ ,则将n称为P的阶若n不存在，则P是无限阶的
    n = 'FFFFFFFEFFFFFFFFFFFFFFFFFFFFFFFF7203DF6B21C6052B53BBF40939D54123'
    a = 'FFFFFFFEFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF00000000FFFFFFFFFFFFFFFC'  # 曲线方程 y^2= x^3+ax+b 的系数a, 16进制字符串
    b = '28E9FA9E9D9F5E344D5A9E4BCF6509A7F39789F515AB8F92DDBCBD414D940E93'  # 曲线方程 y^2= x^3+ax+b 的系数a, 16进制字符串
    p = 'FFFFFFFEFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF00000000FFFFFFFFFFFFFFFF'  # 大于3的一个大素数

    def __repr__(self):
        return '<SM2P256Curve>'

    @property
    def para_len(self) -> int:
        return len(self.n)  # 64

    def scalar_mul(self, k: int, point: str) -> str:
        """
        kP点乘运算 向量乘
        :param k: 倍数k, 整数
        :param point: 曲线点, hex字符串
        :return: 另一个曲线点, hex字符串
        """
        point = '%s%s' % (point, '1')
        mask_str = '8'
        for i in range(self.para_len - 1):
            mask_str += '0'
        mask = int(mask_str, 16)
        temp = point
        flag = False
        for n in range(self.para_len * 4):
            if flag:
                temp = self.double(temp)
            if (k & mask) != 0:
                if flag:
                    temp = self.add(temp, point)
                else:
                    flag = True
                    temp = point
            k = k << 1
        return self._convert_jacb_to_nor(temp)

    def scalar_base_mul(self, k: int) -> str:
        """基点 向量乘"""
        point = ''.join([self.gx, self.gy])
        return self.scalar_mul(k, point)

    def double(self, point: str) -> Union[str, None]:  # TODO Remove this
        """
        倍点
        :param point:
        :return:
        """
        if len(point) < self.para_len * 2:
            return None

        x1, y1, z1 = self._get_point_x_y_z(point)

        p = int(self.p, 16)
        a = int(self.a, 16)

        a3 = a + 3 % p
        t6 = (z1 * z1) % p
        t2 = (y1 * y1) % p
        t3 = (x1 + t6) % p
        t4 = (x1 - t6) % p
        t1 = (t3 * t4) % p
        t3 = (y1 * z1) % p
        t4 = (t2 * 8) % p
        t5 = (x1 * t4) % p
        t1 = (t1 * 3) % p

        t6 = (t6 * t6) % p
        t6 = (a3 * t6) % p
        t1 = (t1 + t6) % p
        z3 = (t3 + t3) % p
        t3 = (t1 * t1) % p
        t2 = (t2 * t4) % p
        x3 = (t3 - t5) % p

        t4 = (t5 + ((t5 + p) >> 1) - t3) % p if (t5 % 2) == 1 else (t5 + (t5 >> 1) - t3) % p
        t1 = (t1 * t4) % p
        y3 = (t1 - t2) % p

        fmt = ('%%0%dx' % self.para_len) * 3
        return fmt % (x3, y3, z3)

    def _get_point_x_y_z(self, point: str) -> (int, int):
        x = int(point[0:self.para_len], 16)
        y = int(point[self.para_len:self.para_len * 2], 16)
        z = 1 if len(point) == self.para_len * 2 else int(point[self.para_len * 2:], 16)
        return x, y, z

    def add(self, point1: str, point2: str) -> Union[str, None]:  # TODO Remove this
        """
        点加函数，P2点为仿射坐标即z=1，P1为Jacobian加重射影坐标
        :param point1:
        :param point2:
        :return:
        """

        if len(point1) < 2 * self.para_len or len(point2) < 2 * self.para_len:
            return None

        x1, y1, z1 = self._get_point_x_y_z(point1)
        x2, y2, _ = self._get_point_x_y_z(point2)

        p = int(self.p, 16)

        t1 = (z1 * z1) % p
        t2 = (y2 * z1) % p
        t3 = (x2 * t1) % p

        t1 = (t1 * t2) % p  # Z1 * Z1 * T2
        t2 = (t3 - x1) % p  # x2 * T1 - X1
        t3 = (t3 + x1) % p  # x2 * T1 + X1
        t4 = (t2 * t2) % p  # (y2 * Z1) * (y2 * Z1)

        t1 = (t1 - y1) % p  # Z1 * Z1 * T2 - Y1
        z3 = (z1 * t2) % p  # Z1 * (x2 * T1 - X1)
        t2 = (t2 * t4) % p  # (x2 * T1 - X1) * (y2 * Z1) * (y2 * Z1)
        t3 = (t3 * t4) % p  # Z1 * (x2 * T1 - X1) * (y2 * Z1) * (y2 * Z1)
        t5 = (t1 * t1) % p  # (Z1 * Z1 * T2 - Y1) * (Z1 * Z1 * T2 - Y1)
        t4 = (x1 * t4) % p  # X1 * (y2 * Z1) * (y2 * Z1)
        x3 = (t5 - t3) % p  # (Z1 * Z1 * T2 - Y1) * (Z1 * Z1 * T2 - Y1) - Z1 * (x2 * T1 - X1) * (y2 * Z1) * (y2 * Z1)
        t2 = (y1 * t2) % p  # Y1 * (x2 * T1 - X1) * (y2 * Z1) * (y2 * Z1)
        t3 = (t4 - x3) % p  # X1 * (y2 * Z1) * (y2 * Z1) - Z1 * (x2 * T1 - X1) * (y2 * Z1) * (y2 * Z1)
        t1 = (t1 * t3) % p  # (Z1 * Z1 * T2 - Y1) * Z1 * (x2 * T1 - X1) * (y2 * Z1) * (y2 * Z1)

        # (Z1 * Z1 * T2 - Y1) * Z1 * (x2 * T1 - X1) * (y2 * Z1) *
        # (y2 * Z1) - Y1 * (x2 * T1 - X1) * (y2 * Z1) * (y2 * Z1)
        y3 = (t1 - t2) % p
        fmt = ('%%0%dx' % self.para_len) * 3
        return fmt % (x3, y3, z3)

    def _convert_jacb_to_nor(self, point: str) -> Union[str, None]:  # Jacobian加重射影坐标转换成仿射坐标
        x, y, z = self._get_point_x_y_z(point)
        p = int(self.p, 16)
        z_inv = pow(z, p - 2, p)
        z_inv_squar = (z_inv * z_inv) % p
        z_inv_qube = (z_inv_squar * z_inv) % p
        x_new = (x * z_inv_squar) % p
        y_new = (y * z_inv_qube) % p
        z_new = (z * z_inv) % p
        if z_new != 1:
            return None
        fmt = ('%%0%dx' % self.para_len) * 2
        return fmt % (x_new, y_new)


sm2p256v1 = SM2P256Curve()


class SM2WithSM3:
    def __init__(self, algorithm=sm3.sm3_hash):
        self._algorithm = algorithm

    @property
    def algorithm(self):
        return self._algorithm


sm2_with_sm3 = SM2WithSM3()


class SM2Signature:
    """sm2签名对象"""

    def __init__(self, r: int, s: int):
        self.r = r
        self.s = s

    def __str__(self):
        return '%064x%064x' % (self.r, self.s)

    def __repr__(self):
        return '<SM2Signature r="%d" s="%d">' % (self.r, self.s)

    @classmethod
    def load(cls, data: bytes) -> "SM2Signature":
        """加载asn1序列化内容构造签名对象"""
        sig, _ = decoder.decode(data)
        r, s = [tag._value for tag in sig.components]
        return cls(r, s)

    def dump(self) -> bytes:
        """按asn1序列化成二进制"""
        seq = SequenceOf(componentType=Integer())
        seq.extend([self.r, self.s])
        return encoder.encode(seq)


class SM2PublicKey:  # TODO ec.point
    """
    公钥 公钥是在椭圆曲线上的一个点，由一对坐标（x，y）组成
    公钥字符串可由 x || y 即 x 拼接 y代表
    """

    def __init__(self, x: str, y: str, curve: SM2P256Curve = sm2p256v1):
        self.curve: SM2P256Curve = curve
        assert curve.is_on_curve(x=hex_to_int(x), y=hex_to_int(y)), '公钥点x,y不在曲线上'
        self.x = x  # 公钥X坐标, 16进制字符串
        self.y = y  # 公钥Y坐标, 16进制字符串

    def __repr__(self):
        return '<SM2PublicKey x="%s" y="%s">' % (self.x, self.y)

    def __eq__(self, other):
        return self.curve == other.curve and self.x == other.x and self.y == other.y

    @staticmethod
    def _calc_entla(uid: bytes) -> str:  # 对勾 ✅
        """
        计算ENTLA, 16进制字符串
        :param uid: uid: 用户A bytes类型的id, 默认为 b'1234567812345678'
        :return: 返回16进制字符串，默认结果为 '0080'
        """
        entla = 8 * len(uid)  # 128
        # 128 >> 8 128(二进制)右移8位, 相当于128 除以 2的8次方, 即 128 // 2 ** 8
        # (整数).to_bytes(1, byteorder='big').hex() 将int转为16进制字符串, 相当于 str(hex(整数))[2:]
        entla1 = (entla >> 8 & 255).to_bytes(1, byteorder='big').hex()
        entla2 = (entla & 255).to_bytes(1, byteorder='big').hex()
        return ''.join([entla1, entla2])  # 拼接entla1 || entla2，相当于 entla1 + entla2

    def _calc_za(self, uid: bytes = DEFAULT_UID) -> str:  # ✅
        """
        使用公钥和用户ID生成ZA-用户身份标识
        ZA=h256(ENTLA || IDA || a || b || G || x || y)  其中G为 Gx || Gy
        :param x: 公钥x坐标, 16进制字符串
        :param y: 公钥y坐标, 16进制字符串
        :param uid: 用户id, bytes字符串, 默认为 b'1234567812345678'
        :return:
        """
        entla = self._calc_entla(uid)  # '0080'
        ida = uid.hex()  # '31323334353637383132333435363738'
        a, b, gx, gy, x, y = self.curve.a, self.curve.b, self.curve.gx, self.curve.gy, self.x, self.y
        z = h256(entla, ida, a, b, gx, gy, x, y)
        za = sm3.sm3_hash(z)
        return za

    def _get_digest(self, data: bytes, uid: bytes = DEFAULT_UID, hash_algorithm=sm3.sm3_hash) -> str:
        """
        通过SM3哈希算法计算消息摘要
        :param hash_algorithm:
        :param data: 消息数据, bytes类型
        :param uid: 用户id, bytes字符串, 默认为 b'1234567812345678'
        :return:
        """
        # 杂凑
        za = self._calc_za(uid)
        data_with_za = bytes.fromhex(za + data.hex())  # 待签名消息
        digest = hash_algorithm(data_with_za)
        return digest

    def encrypt(self, data: bytes, k: str = None, mode=1):
        """
        公钥加密函数
        :param data: 待加密消息
        :param k: 16进制随机数
        :param mode:
        :return:
        """
        # 消息转化为16进制字符串
        data = data.hex()

        para_len = self.curve.para_len

        # 生成随机数
        if k is None:
            k = random_hex(para_len)
        random_key_int = hex_to_int(k)

        # 基点坐标G
        g = self.curve.gx + self.curve.gy
        # 公钥坐标P
        pub = self.x + self.y

        c1 = self.curve.scalar_mul(random_key_int, g)
        xy = self.curve.scalar_mul(random_key_int, pub)
        x2, y2 = xy[0:para_len], xy[para_len:2 * para_len]

        # 消息长度
        data_length = len(data)

        t = sm3.sm3_kdf(xy.encode('utf8'), data_length // 2)
        if int(t, 16) == 0:
            return None
        else:
            form = '%%0%dx' % data_length
            # C2 为密文，与明文长度等长
            c2 = form % (int(data, 16) ^ int(t, 16))
            # C3 为 SM3 算法对明文数计算得到消息摘要，长度固定为 256 位
            c3 = sm3.sm3_hash([i for i in hex_to_bytes(x2, data, y2)])
            if mode:
                return hex_to_bytes(c1, c3, c2)
            else:
                return hex_to_bytes(c1, c2, c3)

    def _verify(self, sig: SM2Signature, data_digest: bytes):
        """
        验证签名及消息摘要
        :param sig: 签名对象
        :param data_digest: 消息摘要
        :return:
        """
        # 验签函数，sign签名r||s，E消息hash，public_key公钥
        r, s = sig.r, sig.s

        # 消息转化为16进制字符串
        e = hex_to_int(data_digest.hex())
        n = hex_to_int(self.curve.n)

        t = (r + s) % n
        if t == 0:
            return 0

        point_g = ''.join([self.curve.gx, self.curve.gy])
        point_public_key = self.x + self.y
        point1 = self.curve.scalar_mul(s, point_g)
        point2 = self.curve.scalar_mul(t, point_public_key)
        if point1 == point2:
            point1 = '%s%s' % (point1, 1)
            point1 = self.curve.double(point1)
        else:
            point1 = '%s%s' % (point1, 1)
            point1 = self.curve.add(point1, point2)
            point1 = self.curve._convert_jacb_to_nor(point1)

        x = int(point1[0:self.curve.para_len], 16)
        return r == (e + x) % n

    def verify_with_sm3(self, sig: SM2Signature, data: bytes, uid: bytes = DEFAULT_UID) -> bool:
        """
        签名验证
        :param sig: 签名对象
        :param data: 原始消息
        :param uid: 用户UID
        :return: 验证成功返回True，否则返回False
        """
        digest = self._get_digest(data, uid)  # 消息摘要
        sign_data = binascii.a2b_hex(digest.encode('utf-8'))
        return self._verify(sig, sign_data)

    def verify(self, signature: bytes, data: bytes, uid: bytes = DEFAULT_UID,
               signature_algorithm=sm2_with_sm3):
        sig = SM2Signature.load(signature)
        digest = self._get_digest(data, uid, hash_algorithm=signature_algorithm.algorithm)  # 消息摘要
        return self._verify(sig, digest.encode('utf-8'))

    def public_bytes(self, encoding='PEM', format='PKCS#8') -> bytes:
        return self.to_pem()

    @classmethod
    def from_pem(cls, pem_bytes) -> "SM2PublicKey":
        pem_bytes = pem_bytes.rstrip(b'\n')
        body = b'\n'.join(pem_bytes.split(b'\n')[1:-1])  # 去掉前后两行
        body_bytes = base64.b64decode(body)
        # der, _ = decoder.decode(body_bytes)
        # value = int_to_hex(der.components[1]._value)
        # value = value[2:]  # 去掉开头的04
        # x, y = value[:64], value[64:]
        # return cls(x=x, y=y)
        return cls.load(body_bytes)

    @classmethod
    def load(cls, data: bytes) -> "SM2PublicKey":
        public_key, _ = decoder.decode(data, asn1Spec=sm2_asn1.SM2PublicKey())
        value = ''.join('%.2x' % x for x in public_key['publicKey'].asNumbers())
        # 去掉开头的04
        value = value[2:]
        x, y = value[:64], value[64:]
        return cls(x=x, y=y)

    def dump(self) -> bytes:
        pkcs8_key = sm2_asn1.SM2PublicKey()
        pkcs8_key["algorithm"] = sm2_asn1.sm2_algorithm
        value = '04%s%s' % (self.x, self.y)
        pkcs8_key["publicKey"] = BitString(hexValue=value)
        der = encoder.encode(pkcs8_key)
        return der

    def to_pem(self) -> bytes:
        der = self.dump()
        body = base64.b64encode(der)
        lines = [b'-----BEGIN PUBLIC KEY-----']
        lines.extend([body[i:i + PARA_LEN] for i in range(0, len(body), PARA_LEN)])
        lines.append(b'-----END PUBLIC KEY-----\n')
        return b'\n'.join(lines)


class SM2PrivateKey:
    """私钥"""

    def __init__(self, value: str):
        """
        :param value: 私钥值-secret, hex string
        """
        # 私钥本质上就是一个256位的随机整数, 16进制字符串, 私钥可以由d表示
        self.value = value  # int(value, 16) < n
        # 公钥对象
        self._public_key = self._get_public_key()

        self.para_len = len(self._public_key.curve.n)  # 64

    def __repr__(self):
        return '<SM2PrivateKey "%s">' % self.value

    def public_key(self) -> SM2PublicKey:
        """
        公钥
        :return: 公钥对象
        """
        return self._public_key

    def _get_public_key(self, curve=sm2p256v1):
        """
        根据曲线及私钥数字d通过kG计算出公钥
        :param curve:
        :return:
        """
        point_g = ec.Point(curve, hex_to_int(curve.gx), hex_to_int(curve.gy))
        private_key_value = hex_to_int(self.value)  # 私钥
        point_public_key = point_g * private_key_value  # 公钥
        return SM2PublicKey(int_to_hex(point_public_key.x), int_to_hex(point_public_key.y), curve=curve)

    def _sign(self, data_digest: bytes, random_key: str) -> (int, int):
        """
        签名函数
        :param data_digest: 待签名消息摘要（哈希后待消息）
        :param random_key: 随机数, hex string
        :return: 签名待r,s
        """
        # 消息, 私钥数字D, 随机数K, 曲线阶N 转int
        curve = self._public_key.curve

        e, d, random_key, n = map(hex_to_int, [data_digest.hex(), self.value, random_key, curve.n])

        # kg运算
        point1 = curve.scalar_base_mul(random_key)

        x = int(point1[0:self.para_len], 16)

        r = ((e + x) % n)

        if r == 0 or r + random_key == n:
            return None, None

        # 计算 （私钥+1) ^ （N - 2）% N
        d_1 = pow(d + 1, n - 2, n)

        # ((私钥+1) * (随机数 + r) - r )  % N
        s = (d_1 * (random_key + r) - r) % n

        if s == 0:
            return None, None

        return r, s

    def sign_with_sm3(self, data: bytes, random_key: str = None, uid: bytes = DEFAULT_UID) -> SM2Signature:
        """
        签名
        :param data: 待签名消息
        :param random_key: 随机数, hex string
        :param uid:
        :return:
        """
        digest_hex = self._public_key._get_digest(data, uid)  # 消息摘要
        data_digest = binascii.a2b_hex(digest_hex.encode('utf-8'))
        if random_key is None:
            random_key = random_hex(self.para_len)
        r, s = self._sign(data_digest, random_key)  # 16进制
        return SM2Signature(r, s)

    def sign(self, data: bytes, random_key: str = None, uid: bytes = DEFAULT_UID, hash_type='SM3') -> bytes:
        return self.sign_with_sm3(data, random_key, uid).dump()

    def decrypt(self, data: bytes, mode=1):
        """
        解密函数
        :param data: 密文
        :param mode: mode: 0-C1C2C3, 1-C1C3C2 (default is 1)
        :return: None
        """
        curve = self._public_key.curve
        para_len = curve.para_len

        data = data.hex()
        len_2 = 2 * para_len
        len_3 = len_2 + 64

        # C1 为随机产生的公钥
        C1 = data[0:len_2]
        if mode == 1:  # C1C3C2
            # C2 为密文，与明文长度等长
            C2 = data[len_3:]
        else:  # C1C2C3
            C2 = data[len_2:-para_len]

        xy = curve.scalar_mul(int(self.value, 16), C1)

        cl = len(C2)
        t = sm3.sm3_kdf(xy.encode('utf8'), cl // 2)
        if int(t, 16) == 0:
            return None
        else:
            form = '%%0%dx' % cl
            M = form % (int(C2, 16) ^ int(t, 16))
            return bytes.fromhex(M)

    @classmethod
    def from_pem(cls, pem_bytes: bytes):
        """
        从PEM加载私钥
        :param pem_bytes: 私钥PEM二进制
        :return: SM2PrivateKey对象
        """
        pem_bytes = pem_bytes.rstrip(b'\n')
        body = b'\n'.join(pem_bytes.split(b'\n')[1:-1])  # 去掉前后两行
        data = base64.b64decode(body)
        # der, _ = decoder.decode(body_bytes)
        # secret = der.components[2]._value.hex()
        # return cls(secret)
        return cls.load(data)

    @classmethod
    def load(cls, data: bytes) -> "SM2PrivateKey":
        private_key, _ = decoder.decode(data, asn1Spec=sm2_asn1.SM2PrivateKey())
        value = ''.join('%.2x' % x for x in private_key['privateKey'].asNumbers())
        return cls(value=value)

    def dump(self) -> bytes:
        pkcs1_key = OctetString(hexValue=self.value)

        pkcs8_key = sm2_asn1.SM2PrivateKey()
        pkcs8_key["version"] = 0
        pkcs8_key["privateKeyAlgorithm"] = sm2_asn1.sm2_algorithm
        pkcs8_key["privateKey"] = pkcs1_key
        der = encoder.encode(pkcs8_key)
        return der

    def to_pem(self) -> bytes:
        der = self.dump()
        body = base64.b64encode(der)
        lines = [b'-----BEGIN PRIVATE KEY-----']
        lines.extend([body[i:i + PARA_LEN] for i in range(0, len(body), PARA_LEN)])
        lines.append(b'-----END PRIVATE KEY-----\n')
        return b'\n'.join(lines)


def generate_private_key(curve=sm2p256v1, backend: typing.Any = None) -> SM2PrivateKey:
    pass


def derive_private_key(private_value: int, curve=sm2p256v1, backend: typing.Any = None) -> SM2PrivateKey:
    pass



from pyasn1.type import namedtype, univ

SM2_OID = '1.2.840.10045.2.1'
EC_PUBLIC_KEY_OID = '1.2.156.10197.1.301'


class AlgorithmIdentifier(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('algorithm', univ.ObjectIdentifier()),
        namedtype.NamedType('namedCurve', univ.ObjectIdentifier())
    )


class SM2PublicKey(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('algorithm', AlgorithmIdentifier()),
        namedtype.NamedType('publicKey', univ.BitString())
    )


# pkcs8
class SM2PrivateKey(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('version', univ.Integer()),
        namedtype.NamedType('privateKeyAlgorithm', AlgorithmIdentifier()),
        namedtype.NamedType('privateKey', univ.OctetString()),
    )


class SM2Signature(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('r', univ.Integer()),
        namedtype.NamedType('s', univ.Integer()),
    )


sm2_algorithm = AlgorithmIdentifier()
sm2_algorithm["algorithm"] = SM2_OID
sm2_algorithm["namedCurve"] = EC_PUBLIC_KEY_OID



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @FileName     :   sm3.py
# @Function     :   SM4算法实现
import copy

from .utils import xor, rotl, get_uint32_be, put_uint32_be, \
    bytes_to_list, list_to_bytes, padding, unpadding

# Expanded SM4 box table
SM4_BOXES_TABLE = [
    0xd6, 0x90, 0xe9, 0xfe, 0xcc, 0xe1, 0x3d, 0xb7, 0x16, 0xb6, 0x14, 0xc2, 0x28, 0xfb, 0x2c,
    0x05, 0x2b, 0x67, 0x9a, 0x76, 0x2a, 0xbe, 0x04, 0xc3, 0xaa, 0x44, 0x13, 0x26, 0x49, 0x86,
    0x06, 0x99, 0x9c, 0x42, 0x50, 0xf4, 0x91, 0xef, 0x98, 0x7a, 0x33, 0x54, 0x0b, 0x43, 0xed,
    0xcf, 0xac, 0x62, 0xe4, 0xb3, 0x1c, 0xa9, 0xc9, 0x08, 0xe8, 0x95, 0x80, 0xdf, 0x94, 0xfa,
    0x75, 0x8f, 0x3f, 0xa6, 0x47, 0x07, 0xa7, 0xfc, 0xf3, 0x73, 0x17, 0xba, 0x83, 0x59, 0x3c,
    0x19, 0xe6, 0x85, 0x4f, 0xa8, 0x68, 0x6b, 0x81, 0xb2, 0x71, 0x64, 0xda, 0x8b, 0xf8, 0xeb,
    0x0f, 0x4b, 0x70, 0x56, 0x9d, 0x35, 0x1e, 0x24, 0x0e, 0x5e, 0x63, 0x58, 0xd1, 0xa2, 0x25,
    0x22, 0x7c, 0x3b, 0x01, 0x21, 0x78, 0x87, 0xd4, 0x00, 0x46, 0x57, 0x9f, 0xd3, 0x27, 0x52,
    0x4c, 0x36, 0x02, 0xe7, 0xa0, 0xc4, 0xc8, 0x9e, 0xea, 0xbf, 0x8a, 0xd2, 0x40, 0xc7, 0x38,
    0xb5, 0xa3, 0xf7, 0xf2, 0xce, 0xf9, 0x61, 0x15, 0xa1, 0xe0, 0xae, 0x5d, 0xa4, 0x9b, 0x34,
    0x1a, 0x55, 0xad, 0x93, 0x32, 0x30, 0xf5, 0x8c, 0xb1, 0xe3, 0x1d, 0xf6, 0xe2, 0x2e, 0x82,
    0x66, 0xca, 0x60, 0xc0, 0x29, 0x23, 0xab, 0x0d, 0x53, 0x4e, 0x6f, 0xd5, 0xdb, 0x37, 0x45,
    0xde, 0xfd, 0x8e, 0x2f, 0x03, 0xff, 0x6a, 0x72, 0x6d, 0x6c, 0x5b, 0x51, 0x8d, 0x1b, 0xaf,
    0x92, 0xbb, 0xdd, 0xbc, 0x7f, 0x11, 0xd9, 0x5c, 0x41, 0x1f, 0x10, 0x5a, 0xd8, 0x0a, 0xc1,
    0x31, 0x88, 0xa5, 0xcd, 0x7b, 0xbd, 0x2d, 0x74, 0xd0, 0x12, 0xb8, 0xe5, 0xb4, 0xb0, 0x89,
    0x69, 0x97, 0x4a, 0x0c, 0x96, 0x77, 0x7e, 0x65, 0xb9, 0xf1, 0x09, 0xc5, 0x6e, 0xc6, 0x84,
    0x18, 0xf0, 0x7d, 0xec, 0x3a, 0xdc, 0x4d, 0x20, 0x79, 0xee, 0x5f, 0x3e, 0xd7, 0xcb, 0x39,
    0x48,
]

# System parameter
SM4_FK = [0xa3b1bac6, 0x56aa3350, 0x677d9197, 0xb27022dc]

# fixed parameter
SM4_CK = [
    0x00070e15, 0x1c232a31, 0x383f464d, 0x545b6269,
    0x70777e85, 0x8c939aa1, 0xa8afb6bd, 0xc4cbd2d9,
    0xe0e7eef5, 0xfc030a11, 0x181f262d, 0x343b4249,
    0x50575e65, 0x6c737a81, 0x888f969d, 0xa4abb2b9,
    0xc0c7ced5, 0xdce3eaf1, 0xf8ff060d, 0x141b2229,
    0x30373e45, 0x4c535a61, 0x686f767d, 0x848b9299,
    0xa0a7aeb5, 0xbcc3cad1, 0xd8dfe6ed, 0xf4fb0209,
    0x10171e25, 0x2c333a41, 0x484f565d, 0x646b7279
]

SM4_ENCRYPT = 0
SM4_DECRYPT = 1


class CryptSM4(object):

    def __init__(self, mode=SM4_ENCRYPT):
        self.sk = [0] * 32
        self.mode = mode

    # Calculating round encryption key.
    # args:    [in] a: a is a 32 bits unsigned value;
    # return: sk[i]: i{0,1,2,3,...31}.
    @classmethod
    def _round_key(cls, ka):
        b = [0, 0, 0, 0]
        a = put_uint32_be(ka)
        b[0] = SM4_BOXES_TABLE[a[0]]
        b[1] = SM4_BOXES_TABLE[a[1]]
        b[2] = SM4_BOXES_TABLE[a[2]]
        b[3] = SM4_BOXES_TABLE[a[3]]
        bb = get_uint32_be(b[0:4])
        rk = bb ^ (rotl(bb, 13)) ^ (rotl(bb, 23))
        return rk

    # Calculating and getting encryption/decryption contents.
    # args:    [in] x0: original contents;
    # args:    [in] x1: original contents;
    # args:    [in] x2: original contents;
    # args:    [in] x3: original contents;
    # args:    [in] rk: encryption/decryption key;
    # return the contents of encryption/decryption contents.
    @classmethod
    def _f(cls, x0, x1, x2, x3, rk):
        # "T algorithm" == "L algorithm" + "t algorithm".
        # args:    [in] a: a is a 32 bits unsigned value;
        # return: c: c is calculated with line algorithm "L" and nonline algorithm "t"
        def _sm4_l_t(ka):
            b = [0, 0, 0, 0]
            a = put_uint32_be(ka)
            b[0] = SM4_BOXES_TABLE[a[0]]
            b[1] = SM4_BOXES_TABLE[a[1]]
            b[2] = SM4_BOXES_TABLE[a[2]]
            b[3] = SM4_BOXES_TABLE[a[3]]
            bb = get_uint32_be(b[0:4])
            c = bb ^ (rotl(bb, 2)) ^ (rotl(bb, 10)) ^ (rotl(bb, 18)) ^ (rotl(bb, 24))
            return c

        return x0 ^ _sm4_l_t(x1 ^ x2 ^ x3 ^ rk)

    def set_key(self, key, mode):
        key = bytes_to_list(key)
        MK = [0, 0, 0, 0]
        k = [0] * 36
        MK[0] = get_uint32_be(key[0:4])
        MK[1] = get_uint32_be(key[4:8])
        MK[2] = get_uint32_be(key[8:12])
        MK[3] = get_uint32_be(key[12:16])
        k[0:4] = xor(MK[0:4], SM4_FK[0:4])
        for i in range(32):
            k[i + 4] = k[i] ^ (
                self._round_key(k[i + 1] ^ k[i + 2] ^ k[i + 3] ^ SM4_CK[i]))
            self.sk[i] = k[i + 4]
        self.mode = mode
        if mode == SM4_DECRYPT:
            for idx in range(16):
                t = self.sk[idx]
                self.sk[idx] = self.sk[31 - idx]
                self.sk[31 - idx] = t

    def one_round(self, sk, in_put):
        out_put = []
        ulbuf = [0] * 36
        ulbuf[0] = get_uint32_be(in_put[0:4])
        ulbuf[1] = get_uint32_be(in_put[4:8])
        ulbuf[2] = get_uint32_be(in_put[8:12])
        ulbuf[3] = get_uint32_be(in_put[12:16])
        for idx in range(32):
            ulbuf[idx + 4] = self._f(ulbuf[idx], ulbuf[idx + 1], ulbuf[idx + 2], ulbuf[idx + 3], sk[idx])

        out_put += put_uint32_be(ulbuf[35])
        out_put += put_uint32_be(ulbuf[34])
        out_put += put_uint32_be(ulbuf[33])
        out_put += put_uint32_be(ulbuf[32])
        return out_put

    def crypt_ecb(self, input_data):
        # SM4-ECB block encryption/decryption
        input_data = bytes_to_list(input_data)
        if self.mode == SM4_ENCRYPT:
            input_data = padding(input_data)
        length = len(input_data)
        i = 0
        output_data = []
        while length > 0:
            output_data += self.one_round(self.sk, input_data[i:i + 16])
            i += 16
            length -= 16
        if self.mode == SM4_DECRYPT:
            return list_to_bytes(unpadding(output_data))
        return list_to_bytes(output_data)

    def crypt_cbc(self, iv, input_data):
        # SM4-CBC buffer encryption/decryption
        i = 0
        output_data = []
        tmp_input = [0] * 16
        iv = bytes_to_list(iv)
        if self.mode == SM4_ENCRYPT:
            input_data = padding(bytes_to_list(input_data))
            length = len(input_data)
            while length > 0:
                tmp_input[0:16] = xor(input_data[i:i + 16], iv[0:16])
                output_data += self.one_round(self.sk, tmp_input[0:16])
                iv = copy.deepcopy(output_data[i:i + 16])
                i += 16
                length -= 16
            return list_to_bytes(output_data)
        else:
            length = len(input_data)
            while length > 0:
                output_data += self.one_round(self.sk, input_data[i:i + 16])
                output_data[i:i + 16] = xor(output_data[i:i + 16], iv[0:16])
                iv = copy.deepcopy(input_data[i:i + 16])
                i += 16
                length -= 16
            return list_to_bytes(unpadding(output_data))



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

from sm_crypto import __version__, __author__, __email__

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['pyasn1']

test_requirements = ['pytest>=3', ]

setup(
    name='sm-crypto',
    version=__version__,
    author=__author__,
    author_email=__email__,
    license="MIT license",
    description="china gm cryptography such as sm2,sm3,sm4",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    url='https://github.com/hanzhichao/sm-crypto',
    install_requires=requirements,
    include_package_data=True,
    keywords=['sm-crypto', 'sm2', 'gm', 'sm3', 'sm4', 'gmssl'],
    packages=find_packages(include=['sm_crypto', 'sm_crypto.*']),
    test_suite='tests',
    tests_require=test_requirements,
    zip_safe=False,
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)
